import abc
from toon.util import mono_clock


def prevent_if_remote(func):
    """Decorator to raise ValueError in order to prevent accidental use
    of a remote device locally.
    """
    def wrap_if_remote(*args, **kwargs):
        self = args[0]
        if self._local:
            return func(*args, **kwargs)
        raise ValueError('Device is being used on a remote process.')
    return wrap_if_remote


class BaseDevice(metaclass=abc.ABCMeta):
    """Abstract base class for input devices.
    Attributes
    ----------
    sampling_frequency: int
        Expected sampling frequency of the device, used by toon.input.MpDevice for preallocation.
        We preallocate for 1 second of data (e.g. 500 samples for a sampling_frequency of 500 Hz).

    Notes
    -----
    The user supplies `enter` and `exit`, not the dunder methods (`__enter__`, `__exit__`).
    """
    sampling_frequency = 500
    shape = (1,)  # shape can technically be None, but
    ctype = None  # ctype must be present

    def __init__(self, clock=mono_clock.get_time):
        """Create new device.
        Parameters
        ----------
        clock: function or method
            The clock used for timestamping events. Defaults to toon.input.mono_clock, which
            allows us to share a reference time between the parent and child processes. The 
            mono_clock is based off psychopy.clock.MonotonicClock 
            (on Windows, time.perf_counter seems to be relative to when the process is created, 
            which makes it difficult to relate the time between processes).

        Notes
        -----

        Do not start communicating with the device in `__init__`, wait until `enter()`.
        """
        self._local = True  # MpDevice toggles this in the main process
        self.clock = clock

    @abc.abstractmethod
    def read(self):
        pass

    def enter(self):
        # we separate enter and exit from the dunder methods
        # so inheriting classes need not use the decorator
        pass

    def exit(self):
        pass

    @prevent_if_remote
    def __enter__(self):
        self.enter()
        return self

    @prevent_if_remote
    def __exit__(self, *args):
        self.exit()

    @property
    def local(self):
        return self._local

    @local.setter
    def local(self, val):
        self._local = bool(val)



import ctypes
import multiprocessing as mp
import os
from collections import namedtuple
from sys import platform

import numpy as np
from numpy.ctypeslib import as_ctypes_type
from psutil import pid_exists

from toon.input._tbprocess import Process
from toon.util import priority

ret = namedtuple('mpdata', ['time', 'data'])
noneret = ret(None, None)


def shared_to_numpy(mp_arr, dims):
    """Convert a :class:`multiprocessing.Array` to a numpy array.
    Helper function to allow use of a :class:`multiprocessing.Array` as a numpy array.
    Derived from the answer at:
    <https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory-for-multiprocessing>
    """
    return np.frombuffer(mp_arr, dtype=mp_arr._type_).reshape(dims)


class MpDevice(object):
    """Creates and manages a process for polling an input device."""

    def __init__(self, device, buffer_len=None, copy_read=True):
        """Create a new MpDevice.

        Parameters
        ----------
        device: object (derived from toon.input.BaseDevice)
            Input device object.
        buffer_len: int, optional
            Overrides the device's sampling_frequency when specifing the size of the
            circular buffer.
        copy_read: bool, optional
            Copy data returned by `read()` (True, default) or return views (faster, but
            more error-prone).
        """
        self.device = device
        self.buffer_len = buffer_len
        self._copy = copy_read
        self.process = None

    def start(self):
        """Start polling from the device on the child process.
        Allocates all resources and creates the child process.

        Notes
        -----
        Prefer using as a context manager over explicitly starting and stopping.

        Raises
        ------
        Will raise an exception if something goes wrong during instantiation
        of the device.
        """
        # For Macs, use spawn (interaction with OpenGL or ??)
        # Windows only does spawn
        if platform in ['darwin', 'win32']:
            mp.freeze_support()  # for Windows executable support
            try:
                mp.set_start_method('spawn')
            except (AttributeError, RuntimeError):
                pass  # already started a process, or on python2

        n_buffers = 2
        # make one lock per buffer
        self.shared_locks = [mp.Lock() for i in range(n_buffers)]
        self.remote_ready = mp.Event()  # signal to main process that remote is done setup
        self.kill_remote = mp.Event()  # signal to remote process to die
        self.current_buffer_index = mp.Value(ctypes.c_bool, 0, lock=False)

        # figure out number of observations to save between reads
        nrow = 100  # default (100 Hz)
        # if we have a sampling_frequency, allocate 1s worth
        # should be enough wiggle room for 60Hz refresh rate
        if self.device.sampling_frequency:
            nrow = self.device.sampling_frequency
        if self.buffer_len:  # buffer_len overcomes all
            nrow = self.buffer_len
        nrow = int(max(nrow, 1))  # make sure we have at least one row

        # preallocate data
        # we have a double buffer sort of thing going on,
        # so we need two of everything
        self._data = []
        time_type = as_ctypes_type(type(self.device.clock()))
        for lck in self.shared_locks:
            # make mp version
            is_scalar = False
            if self.device.shape == (1,):
                new_dim = (nrow,)
                is_scalar = True
            else:
                new_dim = (nrow,) + self.device.shape
            flat_dim = int(np.product(new_dim))
            ctype = self.device.ctype
            # Structures get padding when passing through this,
            # so only run on non-Structures
            if not issubclass(ctype, ctypes.Structure):
                ctype = as_ctypes_type(ctype)
            mp_arr = mp.Array(ctype, flat_dim, lock=False)
            np_arr = shared_to_numpy(mp_arr, new_dim)
            t_mp_arr = mp.Array(time_type, nrow, lock=False)
            t_np_arr = shared_to_numpy(t_mp_arr, nrow)
            counter = mp.Value(ctypes.c_uint, 0, lock=False)
            # generate local version
            local_arr = np.empty_like(np_arr)
            t_local_arr = np.empty_like(t_np_arr)
            if is_scalar:
                local_arr = np.squeeze(local_arr)
            # special case for buffer size of 1 and scalar data
            if local_arr.shape == ():
                local_arr.shape = (1,)
            data_pack = {'mp_data': mp_arr, 'np_data': np_arr,
                         'mp_time': t_mp_arr, 'np_time': t_np_arr,
                         'counter': counter, 'lock': lck}
            self._data.append(data_pack)

        # make local versions to copy the data into
        self._local_arr = local_arr
        self._t_local_arr = t_local_arr
        self.device.local = True

        # TODO: handle errors (current way is broken by python)
        self.process = Process(target=remote,
                               kwargs={'dev': self.device,
                                       'data': self._data,
                                       'remote_ready': self.remote_ready,
                                       'kill_remote': self.kill_remote,
                                       'parent_pid': os.getpid(),
                                       'current_buffer_index': self.current_buffer_index})

        self.process.daemon = True
        self.process.start()
        self.check_error()
        self.remote_ready.wait()  # block until child process is ready
        self.device.local = False  # try to prevent local access to the device

    def read(self):
        # TODO: add docs
        self.check_error()
        # get the current buffer (either 0 or 1)
        current_buffer_index = self.current_buffer_index.value
        # this might block, if the remote is currently writing data
        current_data = self._data[current_buffer_index]
        with current_data['lock']:
            local_count = current_data['counter'].value
            current_data['counter'].value = 0  # start writing from the top of the array
            if local_count > 0:
                t_out = self._t_local_arr[:local_count]
                data_out = self._local_arr[:local_count]
                data_out[:] = current_data['np_data'][:local_count]
                t_out[:] = current_data['np_time'][:local_count]
        if local_count <= 0:
            return None
        # return time, data
        if self._copy:
            return ret(np.copy(t_out), np.copy(data_out))
        # otherwise, return views
        return ret(t_out, data_out)

    def clear(self):
        """Discard all pending observations."""
        self.check_error()
        current_buffer_index = self.current_buffer_index.value
        current_data = self._data[current_buffer_index]
        with current_data['lock']:
            current_data['counter'].value = 0

    def check_error(self):
        if self.process:
            if not self.process.is_alive():
                if self.process.exception:
                    err, traceback = self.process.exception
                    print(traceback)
                    raise err
                else:
                    raise ValueError('MpDevice is closed.')

    def stop(self):
        self.kill_remote.set()
        self.process.join()
        self.device.local = True

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()


def process_data(shared_time, shared_data, local_time, local_data, shared_counter):
    next_index = shared_counter.value
    if next_index < shared_time.shape[0]:
        shared_time[next_index] = local_time
        shared_data[next_index] = local_data
        shared_counter.value += 1
    else:  # ring buffer ish, see benchmarks https://github.com/aforren1/toon/issues/77
        shared_time[:-1] = shared_time[1:]
        shared_time[-1] = local_time
        shared_data[:-1] = shared_data[1:]
        shared_data[-1] = local_data


def remote(dev, data, remote_ready, kill_remote, parent_pid, current_buffer_index):
    # from timeit import default_timer
    for d in data:
        # need to re-generate connection between mp and np arrays
        dims = d['np_data'].shape
        d['np_data'] = shared_to_numpy(d['mp_data'], dims)
        d['np_time'] = shared_to_numpy(d['mp_time'], dims[0])
    try:
        priority(1)  # high priority (non-realtime, though) and disables gc
        with dev:
            remote_ready.set()  # signal all set to the parent process
            while not kill_remote.is_set() and pid_exists(parent_pid):
                # either a (time, data) tuple or list of (time, data) tuples
                # or None if nothing
                device_dat = dev.read()
                # t0 = default_timer()
                if device_dat is None:
                    continue  # next read
                buffer_index = current_buffer_index.value
                # lock magicks
                # test whether the current buffer is accessible
                current_data = data[buffer_index]
                lck = current_data['lock']
                success = lck.acquire(block=False)
                if not success:
                    current_buffer_index.value = not current_buffer_index.value
                    buffer_index = not buffer_index
                    current_data = data[buffer_index]
                    lck = current_data['lock']
                    # manual lock handling
                    lck.acquire()
                try:
                    shared_time = current_data['np_time']
                    shared_data = current_data['np_data']
                    shared_counter = current_data['counter']
                    if isinstance(device_dat, list):
                        for dat in device_dat:
                            process_data(shared_time, shared_data,
                                         dat[0], dat[1], shared_counter)
                    else:
                        process_data(shared_time, shared_data,
                                     device_dat[0], device_dat[1], shared_counter)
                finally:
                    lck.release()
                # print(default_timer() - t0)

    finally:
        priority(0)
        remote_ready.set()



# rush derived from Psychtoolbox
import os
import gc
import sys
import warnings


if sys.platform == 'win32':
    kernel32 = None
    avrt = None
    from ctypes import WinDLL, get_last_error, set_last_error, byref
    from ctypes.wintypes import LPDWORD, LPCSTR
    try:
        kernel32 = WinDLL('kernel32', use_last_error=True)
    except Exception:
        warnings.warn('kernel32 import failed.')
    try:
        avrt = WinDLL('Avrt', use_last_error=True)
    except Exception:
        warnings.warn('Avrt import failed.')

    # https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-setpriorityclass
    NORMAL_PRIORITY_CLASS = 0x00000020
    HIGH_PRIORITY_CLASS = 0x00000080
    REALTIME_PRIORITY_CLASS = 0x00000100

    # https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-setthreadpriority
    THREAD_PRIORITY_NORMAL = 0
    THREAD_PRIORITY_ABOVE_NORMAL = 1
    THREAD_PRIORITY_HIGHEST = 2
    THREAD_PRIORITY_TIME_CRITICAL = 15

    # https://docs.microsoft.com/en-us/windows/win32/procthread/process-security-and-access-rights
    PROCESS_SET_INFORMATION = 0x0200
    PROCESS_QUERY_INFORMATION = 0x0400

    thread_pool = {}  # keep a dict of

    # refs:
    # https://github.com/Psychtoolbox-3/Psychtoolbox-3/blob/e65061d5e7e6ff1fcf45c50eecd5819daa857919/PsychSourceGL/Source/Common/Screen/SCREENGetMouseHelper.c#L608
    # https://github.com/Psychtoolbox-3/Psychtoolbox-3/blob/2ad56475c2b5059701e2616cc1b77a6f6f20380b/PsychSourceGL/Source/Windows/Base/PsychTimeGlue.c#L1358
    def priority(level=0, pid=None):
        # 0 = normal priority
        # 1 = high
        # 2 = realtime
        if level > 0:
            gc.disable()
        else:
            gc.enable()

        if kernel32 is None and avrt is None:
            # nothing to do if we don't have these
            return False

        pid = os.getpid()
        process = kernel32.OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_SET_INFORMATION, 0, pid)
        thread = kernel32.GetCurrentThread()

        if thread in thread_pool.keys():
            thread_grp = thread_pool[thread]
        else:
            thread_grp = {'mmcss': None}
            thread_pool[thread] = thread_grp

        # if this is a MMCSS scheduled thread, need to revert to normal first
        if avrt and thread_grp['mmcss']:
            avrt.AvRevertMmThreadCharacteristics(thread_grp['mmcss'])
            thread_grp['mmcss'] = None

        # clear error if anything?
        err = get_last_error()
        if err:
            set_last_error(0)
            return False
        if level == 0:
            kernel32.SetPriorityClass(process, NORMAL_PRIORITY_CLASS)
            kernel32.SetThreadPriority(thread, THREAD_PRIORITY_NORMAL)
        elif level == 1:
            kernel32.SetPriorityClass(process, HIGH_PRIORITY_CLASS)
            if avrt:
                tmp = LPDWORD()
                thread_grp['mmcss'] = avrt.AvSetMmMaxThreadCharacteristicsA(LPCSTR(b'Pro Audio'), LPCSTR(b'Capture'), byref(tmp))
            if not thread_grp['mmcss']:  # failed
                kernel32.SetThreadPriority(thread, THREAD_PRIORITY_HIGHEST)

        elif level == 2:
            # first try to set time critical
            if (not kernel32.SetPriorityClass(process, REALTIME_PRIORITY_CLASS) or
                    kernel32.GetPriorityClass(process) != REALTIME_PRIORITY_CLASS):
                # try for high priority scheduling instead
                kernel32.SetPriorityClass(process, HIGH_PRIORITY_CLASS)
            if avrt:
                tmp = LPDWORD()
                thread_grp['mmcss'] = avrt.AvSetMmMaxThreadCharacteristicsA(LPCSTR(b'Pro Audio'), LPCSTR(b'Capture'), byref(tmp))
            if not thread_grp['mmcss']:  # failed
                kernel32.SetThreadPriority(thread, THREAD_PRIORITY_ABOVE_NORMAL)

        return True

elif sys.platform == 'darwin':
    # TODO
    # https://developer.apple.com/library/archive/documentation/Darwin/Conceptual/KernelProgramming/scheduler/scheduler.html#//apple_ref/doc/uid/TP30000905-CH211-BABCHEEB
    def priority(level=0, pid=None):
        if level > 0:
            gc.disable()
        else:
            gc.enable()
        return True

else:  # linux
    import os
    import ctypes
    import ctypes.util

    MCL_CURRENT = 1
    MCL_FUTURE = 2

    libc = ctypes.CDLL(ctypes.util.find_library('c'))

    def priority(level=0, pid=0):
        if level > 0:
            gc.disable()
        else:
            gc.enable()

        libc.munlockall()
        if level == 1:
            policy = os.SCHED_RR
        elif level == 2:
            policy = os.SCHED_FIFO
        else:
            policy = os.SCHED_OTHER

        sched_param = os.sched_param(os.sched_get_priority_max(policy))

        try:
            os.sched_setscheduler(pid, policy, sched_param)
        except PermissionError:
            return False

        if level == 2:  # try to lock memory (and we succeeded in sched already)
            res = libc.mlockall(MCL_CURRENT | MCL_FUTURE)
            if res != 0:
                libc.munlockall()

        return True



from pytest import raises
import numpy as np
from tests.input.mockdevices import Dummy, Timebomb, DummyList, SometimesNot, StructObs


def test_device_single():
    dev = Dummy()
    with dev:
        time, data = dev.read()
    assert(isinstance(time, float))
    assert(isinstance(data, np.ndarray))
    assert(data.shape == (5,))


def test_err():
    dev = Timebomb()
    with dev:
        with raises(ValueError):
            for i in range(12):
                dev.read()


def test_list():
    dev = DummyList()
    with dev:
        data = dev.read()

    assert(isinstance(data, list))
    assert(len(data) == 5)


def test_nones():
    dev = SometimesNot()
    data = []
    with dev:
        for i in range(10):
            data.append(dev.read())

    assert(data[1] is None)


def test_struct():
    dev = StructObs()
    with dev:
        time, data = dev.read()
    assert(data.ll.x == 0)



from toon.input.device import BaseDevice
from timeit import default_timer
import ctypes
import numpy as np


class Dummy(BaseDevice):
    t0 = default_timer()
    ctype = ctypes.c_int
    shape = (5,)
    sampling_frequency = 500

    def read(self):
        data = np.random.randint(5, size=self.shape)
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        self.t0 = default_timer()
        t = self.clock()
        return t, data


class NoData(Dummy):
    def read(self):
        return None


class Timebomb(Dummy):
    counter = 0

    def read(self):
        self.counter += 1
        if self.counter > 10:
            raise ValueError('Broke it.')
        return super().read()


class DummyList(Dummy):
    def read(self):
        out = []
        for i in range(5):
            out.append(super().read())
        return out


class SometimesNot(Dummy):
    counter = 0

    def read(self):
        self.counter += 1
        if self.counter % 2 == 0:
            return None
        return super().read()


class Point(ctypes.Structure):
    _fields_ = [('x', ctypes.c_int), ('y', ctypes.c_int)]


class Rect(ctypes.Structure):
    _fields_ = [('ll', Point), ('ur', Point)]


class StructObs(BaseDevice):
    ctype = Rect
    sampling_frequency = 200
    shape = (2, 2)
    t0 = default_timer()

    def read(self):
        data = Rect()
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        self.t0 = default_timer()
        t = self.clock()
        return t, data


class Incrementing(BaseDevice):
    ctype = int
    sampling_frequency = 100
    shape = (1,)
    t0 = default_timer()
    counter = 0

    def read(self):
        data = self.counter
        self.counter += 1
        while default_timer() - self.t0 < (1.0/self.sampling_frequency):
            pass
        self.t0 = default_timer()
        t = self.clock()
        return t, data



"""
  Copyright 2014, 2015, 2016 Ori Livneh <ori@wikimedia.org>
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

# copied from https://github.com/atdt/monotonic
# under the Apache License 2.0
# starting from
# https://github.com/atdt/monotonic/blob/5bd1bf90e5e49d4ce4f5a90b9c97de6934b9503c/monotonic.py#L56
# until the end of the `if sys.platform == 'darwin'` block (~line 76)
# pragma: no cover
import ctypes

libc = ctypes.CDLL('/usr/lib/libc.dylib', use_errno=True)


class mach_timebase_info_data_t(ctypes.Structure):
    _fields_ = (('numer', ctypes.c_uint32), ('denom', ctypes.c_uint32))


mach_absolute_time = libc.mach_absolute_time
mach_absolute_time.restype = ctypes.c_uint64


class MonoClock(object):
    def __init__(self, relative=True):
        self._reference_counter = self.get_ticks() if relative else 0
        timebase = mach_timebase_info_data_t()
        libc.mach_timebase_info(ctypes.byref(timebase))
        self.inv_frequency = 1/(timebase.numer / timebase.denom * 1.0e9)

    def get_ticks(self):
        return mach_absolute_time()

    def get_time(self):
        return (self.get_ticks() - self._reference_counter) * self.inv_frequency

    def getTime(self):
        return self.get_time()

    @property
    def start_time(self):
        return self._reference_counter * self.inv_frequency



import ctypes
import ctypes.wintypes as cwt
kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)


# From Stack Overflow
# User linusg https://stackoverflow.com/users/5952681/linusg
# https://stackoverflow.com/a/38463185/2690232
class MonoClock(object):
    def __init__(self, relative=True):
        self._reference_counter = self.get_ticks() if relative else cwt.LARGE_INTEGER(0)
        self.frequency = cwt.LARGE_INTEGER()
        kernel32.QueryPerformanceFrequency(ctypes.byref(self.frequency))
        self.inv_frequency = 1/float(self.frequency.value)

    def get_ticks(self):
        current_counter = cwt.LARGE_INTEGER()
        kernel32.QueryPerformanceCounter(ctypes.byref(current_counter))
        return current_counter

    def get_time(self):
        return (self.get_ticks().value - self._reference_counter.value) * self.inv_frequency

    def getTime(self):
        return self.get_time()

    @property
    def start_time(self):
        return self._reference_counter.value * self.inv_frequency



from inspect import ismethod
from collections import namedtuple
from copy import copy

TrackAttr = namedtuple('TrackAttr', 'track attr obj kwargs')


class Player(object):
    """Manages one or more Tracks, allowing for synchronization of animations.

    Notes
    -----
    Can also be used as a mixin.
    """

    def __init__(self, repeats=1, *args, **kwargs):
        """Create a Player.

        Parameters
        ----------
        repeats: number of times to repeat the animation
        *args: arbitrary (only matters if using as mixin)
        **kwargs: arbitrary (only matters if using as mixin)
        """
        self.tracks = []
        self.state = 'stopped'  # 'playing'
        self.ref_time = 0
        self.duration = 0
        self.timescale = 1
        if repeats < 1:
            raise ValueError('Number of repeats must be > 1 (or math.inf)')
        self._repeats = repeats  # track
        self.repeats = repeats

    def add(self, track, attr, obj=None, **kwargs):
        """Add a new Track to the Player.

        Parameters
        ----------
        track: toon.anim.Track
            A Track object.
        attr: str or function
            The attribute/property to change, or a function that takes (val, obj)
            as positional arguments, where `val` is the interpolated value, and
            `obj` is the object to be manipulated.
        obj: None (if mixin), object or list of objects (otherwise)
            Object that is being manipulated.
        **kwargs: Arbitrary keyword arguments, passed to `attr` if it is a function
        """
        self.tracks.append(TrackAttr(copy(track), attr, obj, kwargs))
        new_dur = track.duration()
        self.duration = new_dur if new_dur > self.duration else self.duration

    def start(self, time):
        """Start the animations.

        Parameters
        ----------
        time: float
            Reference time.
        """
        self.ref_time = time
        self._repeats = self.repeats
        self.state = 'playing'
        self.advance(time)  # start should reset everything to the initial time?

    def stop(self):
        """Stop the animations."""
        self.state = 'stopped'

    def reset(self):
        """Reset all tracks and stop animation."""
        self.start(self.ref_time)
        self.stop()

    def resume(self, time):
        """Resume playing.

        Notes
        -----
        If already playing, do nothing.
        """
        if self.state == 'playing':
            return
        self.start(time)

    def _do_update(self, attr, val, obj, **kwargs):
        # if we get a function, call function with updated val
        if callable(attr):
            # if it's a method (e.g. a setter), we call that method with the val as an arg
            if ismethod(attr):
                if kwargs:
                    attr(val, **kwargs)
                else:
                    attr(val)
                return
            if kwargs:
                attr(val, obj, **kwargs)
            else:
                attr(val, obj)
            return
        # otherwise (user gave string), directly set the attribute
        setattr(obj, attr, val)

    def advance(self, time):
        """Update all Tracks with a new time.

        Parameters
        ----------
        time: float
            Current time. Track time is `time - reference time`.
        """
        if self.state != 'playing':
            return
        if time < self.ref_time:
            return
        for trk in self.tracks:
            # if tracks are playing, will return a val
            val = trk.track.at((time - self.ref_time) * self.timescale)
            if trk.obj is not None:  # object or list provided, so we'll manipulate them
                try:  # see if single object
                    self._do_update(trk.attr, val,
                                    trk.obj, **trk.kwargs)
                except (TypeError, AttributeError):  # list of objects?
                    for obj in trk.obj:
                        self._do_update(trk.attr, val,
                                        obj, **trk.kwargs)
            else:  # operate on self
                self._do_update(trk.attr, val, self, **trk.kwargs)
            # if we've gone beyond, stop playing
            if time - self.ref_time >= self.duration:
                self._repeats -= 1
                if self._repeats < 1:
                    self.state = 'stopped'
                else:
                    self.ref_time = time

    @property
    def is_playing(self):
        return self.state == 'playing'

    @property
    def is_stopped(self):
        return self.state == 'stopped'

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, val):
        self._state = val

    @property
    def timescale(self):
        return self._timescale

    @timescale.setter
    def timescale(self, value):
        """Scales the animations.

        Parameters
        ----------
        value: float
            Value on the interval (0, Inf], where < 1 slows down the animation,
            and > 1 speeds up the animation.

        Notes
        -----
        Defaults to 1 (normal playback speed).
        """
        self._timescale = value



from time import sleep
from pytest import raises, approx
import numpy as np
from tests.input.mockdevices import (Dummy, Timebomb, DummyList,
                                     SometimesNot, StructObs, Incrementing,
                                     NoData)
from toon.util import mono_clock
from toon.input import MpDevice

Dummy.sampling_frequency = 1000


def test_device_single():
    dev = MpDevice(Dummy())
    dev.start()
    sleep(0.25)
    res = dev.read()
    dev.stop()
    assert(res is not None)
    time, data = res
    print(np.diff(time))
    print(data.shape)
    assert(data.shape[0] > 10)
    assert(data.shape[0] == time.shape[0])


def test_err():
    dev = MpDevice(Timebomb())
    with dev:
        sleep(0.2)
        with raises(ValueError):
            dat = dev.read()


def test_list():
    dev = MpDevice(DummyList())
    with dev:
        sleep(0.2)
        time, data = dev.read()
    assert(data.shape[0] > 10)
    assert(data.shape[0] == time.shape[0])


def test_nones():
    dev = MpDevice(SometimesNot())
    with dev:
        sleep(0.2)
        time, data = dev.read()
    print(data.shape)
    assert(data.shape[0] > 10)
    assert(data.shape[0] == time.shape[0])


def test_struct():
    dev = MpDevice(StructObs())
    with dev:
        sleep(0.2)
        time, data = dev.read()
    assert(data.shape[0] > 10)
    assert(data.shape[0] == time.shape[0])


def test_freq():
    original_fs = Dummy.sampling_frequency
    Dummy.sampling_frequency = 1

    dev = MpDevice(Dummy())
    with dev:
        sleep(4)
        time, val = dev.read()

    assert(time.shape[0] == 1)
    assert(val.shape == (1, 5))
    print(time, val)
    Dummy.sampling_frequency = original_fs


def test_ringbuffer():
    dev = MpDevice(Dummy(), buffer_len=1)
    with dev:
        sleep(1)
        time, val = dev.read()

    assert(time.shape[0] == 1)
    assert(val.shape == (1, 5))
    print(time, val)


def test_mono():
    dev = MpDevice(Incrementing(), buffer_len=10)
    with dev:
        sleep(1)
        time, val = dev.read()

    print(time, val)
    assert(all(np.diff(time) > 0))
    assert(all(np.diff(val) > 0))


def test_read_multi():
    dev = MpDevice(Incrementing())
    with dev:
        times, dats = [], []
        for i in range(500):
            data = dev.read()
            if data:
                times.append(data.time)
                dats.append(data.data)
            sleep(0.016)
    times = np.hstack(times)
    vals = np.hstack(dats)
    print(times)
    print(vals)
    assert(all(np.diff(times)) > 0)
    assert(all(np.diff(vals)) > 0)


def test_restart():
    dev = MpDevice(Dummy())
    with dev:
        sleep(0.2)
        res = dev.read()
    with dev:
        sleep(0.2)
        res2 = dev.read()
    print(res[0])
    print(res2[0])
    assert(res[0].any() and res2[0].any())
    assert(res[0][-1] < res2[0][0])


def test_multi_devs():
    dev1 = MpDevice(Dummy())
    dev2 = MpDevice(Dummy())

    with dev1, dev2:
        sleep(0.2)
        time1, data1 = dev1.read()
        time2, data2 = dev2.read()

    print(time1[-5:], time2[-5:])


def test_none():
    dev = MpDevice(NoData())
    with dev:
        sleep(0.2)
        res = dev.read()

    assert(res is None)


def test_remote_clock():
    dev = MpDevice(Dummy(clock=mono_clock.get_time))
    sleep(0.5)
    with dev:
        sleep(0.1)
        time, data = dev.read()
        tmp = mono_clock.get_time()
    print(tmp - time[-1])
    assert(tmp - time[-1] == approx(1.0/Dummy.sampling_frequency, abs=3e-3))


def test_already_closed():
    dev = MpDevice(Dummy())
    dev.start()
    res = dev.read()
    dev.stop()
    with raises(ValueError):
        dev.read()


def test_no_local():
    local_dev = Dummy()
    dev = MpDevice(local_dev)
    with dev:
        with raises(ValueError):
            with local_dev:
                res = local_dev.read()



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# get requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()

# description for pypi
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    desc = f.read()

setup(
    name='toon',
    version='0.13.1',
    description='Tools for neuroscience experiments',
    long_description=desc,
    long_description_content_type='text/markdown',
    url='https://github.com/aforren1/toon',
    author='Alexander Forrence',
    author_email='aforren1@jhu.edu',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    install_requires=requirements,
    keywords='psychophysics neuroscience input experiment',
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'example_devices'])
)



from platform import system

sys = system()
if sys == 'Windows':
    from toon.util.clocks.win_clock import MonoClock
elif sys == 'Darwin':
    from toon.util.clocks.mac_clock import MonoClock
else:  # anything else uses timeit.default_timer
    from toon.util.clocks.default_clock import MonoClock

mono_clock = MonoClock()
